import os
from pathlib import Path

import pytest
from pytest import fixture

# Prevent pytest from catching exceptions when debugging in vscode so that break on
# exception works correctly (see: https://github.com/pytest-dev/pytest/issues/7409)
if os.getenv("PYTEST_RAISE", "0") == "1":

    @pytest.hookimpl(tryfirst=True)
    def pytest_exception_interact(call):
        raise call.excinfo.value

    @pytest.hookimpl(tryfirst=True)
    def pytest_internalerror(excinfo):
        raise excinfo.value


@fixture
def samples():
    return Path(__file__).parent / "samples"



"""
devices package
===============

Each SwitchDevice derived class in this package adds support for a type
of switching device.
"""

from . import shell_cmd, smart_thing, web_device, web_ui

# __all__ defines the public API for the package.
# Each module also defines its own __all__.
__all__ = [
    "shell_cmd",
    "smart_thing",
    "web_ui",
    "web_device",
    "cisco_ios_poe_switch",
    "kasa_device",
]



"""
Test the full software stack with the backend switch services mocked
out.
"""

from base64 import b64encode
from pathlib import Path
from unittest.mock import patch

from ruamel.yaml import YAML

# import all sublasses of SwitchDevice so ApiSchema sees them
from maaspower.devices.shell_cmd import CommandLine
from maaspower.devices.smart_thing import SmartThing
from maaspower.maasconfig import MaasConfig
from maaspower.webhook import app, load_web_hook

# from unittest.mock import patch


# avoid linter complaints
required_to_find_subclasses = [SmartThing, CommandLine]

userpass = "a_user:a_pass"
encoded_u = b64encode(userpass.encode()).decode()
headers = {"Authorization": f"Basic {encoded_u}"}


def test_web_root():
    """
    Test that the web server responds to its root URL
    """
    with app.test_client() as test_client:
        response = test_client.get("/")
        assert response.status_code == 200
        assert b"MAAS Power Web Hooks Server" == response.data


def test_webhook_cmdline(samples: Path):
    """
    Load a dummy command line controlled switch configuration and
    confirm that we can call it via the webhooks
    """

    dummypath = samples / "dummy_cmd.yaml"
    config_dict = YAML().load(dummypath)

    maas_config = MaasConfig.deserialize(config_dict)
    load_web_hook(maas_config)
    app.testing = True

    with app.test_client() as test_client:
        response = test_client.post("/maaspower/pi1/on", headers=headers)
        assert response.status_code == 200
        response = test_client.post("/maaspower/pi1/off", headers=headers)
        assert response.status_code == 200
        response = test_client.post("/maaspower/pi1/query", headers=headers)
        assert response.data == b"status : running"


def test_regex(tmp_path: Path, samples: Path):
    """
    Load up a config that tests regex in device names and call the webook
    with various matching names
    """

    samplepath = samples / "sampleregex.yaml"
    config_dict = YAML().load(samplepath)

    maas_config = MaasConfig.deserialize(config_dict)
    load_web_hook(maas_config)
    app.testing = True

    with app.test_client() as test_client:
        response = test_client.post("/maaspower/hello1/query", headers=headers)
        assert response.data == b"status : running"
        response = test_client.post("/maaspower/192_168_1_3/query", headers=headers)
        assert response.data == b"status : running"


@patch("maaspower.devices.shell_cmd.CommandLine.execute_command", return_value="power")
def test_substitution(command_line, tmp_path: Path, samples: Path):
    """
    Load up a config that tests regex in device names and call the webook
    with various matching names
    """

    samplepath = samples / "sampleregex.yaml"
    config_dict = YAML().load(samplepath)

    maas_config = MaasConfig.deserialize(config_dict)
    load_web_hook(maas_config)

    with app.test_client() as test_client:
        test_client.post("/maaspower/hello1/query", headers=headers)
        assert command_line.call_args.args[0] == "echo hello1 power"
        test_client.post("/maaspower/192_168_1_3/query", headers=headers)
        assert command_line.call_args.args[0] == "echo 192_168_1_3 power"


@patch("maaspower.devices.shell_cmd.CommandLine.execute_command", return_value="power")
def test_substitution_matches(command_line, samples: Path):
    """
    Load up a config that tests regex in device names and call the webook
    with various matching names, also using a match group to do
    substitutions into the command strings
    """

    samplepath = samples / "sampleregex2.yaml"
    config_dict = YAML().load(samplepath)

    maas_config = MaasConfig.deserialize(config_dict)
    load_web_hook(maas_config)
    app.testing = True

    assert len(maas_config._devices) == 1

    with app.test_client() as test_client:
        test_client.post("/maaspower/raspi1-p1/on", headers=headers)
        assert (
            command_line.call_args.args[0]
            == "uhubctl -a 1 -p 1 # turn on raspi1 (full device name was raspi1-p1)"
        )
        # verify the new match added a new device into the cache
        assert len(maas_config._devices) == 2
        test_client.post("/maaspower/raspi2-p3/query", headers=headers)
        assert command_line.call_args.args[0] == "uhubctl -p 3"
        # verify the new match added a new device into the cache
        assert len(maas_config._devices) == 3
        test_client.post("/maaspower/raspi1-p1/off", headers=headers)
        assert command_line.call_args.args[0] == "uhubctl -a 0 -p 1"
        # verify the previous match did not update the cache
        assert len(maas_config._devices) == 3


# @patch("pysmartthings.SmartThings")
# def test_webhook_smartthings(smarthings, samples: Path):
#     """
#     Load a smartthing controlled switch configuration and
#     confirm that we can call it via the webhooks, mock out the
#     smartthings backend
#     """

#     dummypath = samples / "dummy_cmd.yaml"
#     config_dict = YAML().load(dummypath)

#     maas_config = MaasConfig.deserialize(config_dict)
#     load_web_hook(maas_config)

#     with app.test_client() as test_client:
#         response = test_client.post("/maaspower/nuc1/on", headers=headers)
#         assert response.status_code == 200
#         response = test_client.post("/maaspower/nuc1/off", headers=headers)
#         assert response.status_code == 200
#         response = test_client.post("/maaspower/nuc1/query", headers=headers)
#         assert response.data == b"status : running"



# Web UI playground.
#
# To work out the sequence of elements you need to interact with to
# Web scrape the power control and query from your device use this module
# with ipython (replace URL with your device)
#
#    activate your venv
#    pip install ipython
#    ipython webtestui.py -i http://192.168.0.137/index.cgi
#
# This will run the code below and drop you into a prompt and pop up a Chrome
# window
# Now you can use the (right-click ) inspect in chrome to take a look at
# which HTML elements you need to poke to login, and control PoE ports.
# You will need to find a unique property of the element to identify
# it. This code supports classname, name, partial link text and ID. See
# class FindBy for details
#
# take a look at the example sequences login/enable/disable/logout and
# develop your own for your device. These can then be added to the config
# file for your maaspower setup.

# You need Chrome or Chromium installed for this to work
# Also you need to download the ChromeDriver
#
# Get the chrome driver for selenium from here (pick one that matches
# your Chrome version):
# https://chromedriver.storage.googleapis.com/index.html
# and place it in the same folder as this file

import os
import re
import sys
from enum import Enum
from pathlib import Path
from time import sleep

from selenium import webdriver
from selenium.webdriver.common.by import By
from selenium.webdriver.support import expected_conditions as EC
from selenium.webdriver.support.wait import WebDriverWait

this_dir = Path(__file__).parent
os.chdir(this_dir)

driver = webdriver.Chrome("./chromedriver")

url = sys.argv[1]

driver.get(url)
driver.timeouts._implicit_wait = 20

command_regex = re.compile(r"([^\/]*)\/([^\/]*)\/?([^\/]*)?\/?([^\/]*)?$")
index_regex = re.compile(r"(.*)\[([0-9]*)\]")


class FindBy(Enum):
    id = By.ID
    cls = By.CLASS_NAME
    link = By.PARTIAL_LINK_TEXT
    n = By.NAME
    css = By.CSS_SELECTOR


def process_arguments(by_str: str, value: str) -> tuple[str, str, int | None]:
    # convert our short fieldtype string to the full selenium string
    by = FindBy[by_str].value

    # check for and index [n] at the end of a string. Used for
    # matching > 1 html element and we want to index into the list of matches
    idx_match = index_regex.match(value)
    if idx_match is None:
        index = None
    else:
        value = idx_match.group(1)
        index = int(idx_match.group(2))
    return by, value, index


def click(by_str: str, value: str):
    by, value, index = process_arguments(by_str, value)
    print(f"click on {by} {value} with index {index}")
    elem = WebDriverWait(driver, 10).until(EC.element_to_be_clickable((by, value)))
    if index is not None:
        elements = driver.find_elements(by=by, value=value)
        elem = elements[index]
    elem.click()


def send(by_str: str, value: str, text: str):
    by, value, index = process_arguments(by_str, value)
    print(f"send {text} to {by} {value} with index {index}")
    elem = WebDriverWait(driver, 10).until(EC.presence_of_element_located((by, value)))
    if index:
        elements = driver.find_elements(by=by, value=value)
        elem = elements[index]
    elem.send_keys(text)


def get(by_str: str, value) -> str:
    by, value, index = process_arguments(by_str, value)
    print(f"get field {by} {value} with index {index}")
    elem = WebDriverWait(driver, 10).until(EC.presence_of_element_located((by, value)))
    if index:
        elements = driver.find_elements(by=by, value=value)
        elem = elements[index]
    return elem.text


def sequence(elements: list[str]):
    """
    take a sequence of strings of the form
    command_by_value
    where:
      command is click or send
      by is one of id, classname, link, named
      value is the value for the id, class, partial link text, name of the
        html element to interact with
    when the command is 'send' the next string in the sequence is the text to
    send to the html element
    """
    for element in elements:
        # separate out the elements of the string
        match = command_regex.match(element)
        if match is None:
            raise (ValueError(f"bad command format: {element}"))

        command = match.group(1)

        # invoke the command via selenium
        if command == "click":
            click(match.group(2), match.group(3))
        elif command == "send":
            send(match.group(2), match.group(3), match.group(4))
        elif command == "get":
            click(match.group(2), match.group(3))
        elif command == "delay":
            sleep(float(match.group(2)))


# these example sequences work for the Netgear G308EP

login = ["send/cls/pwd-field-text/MYPASSWORD\n", "click/link/POE"]

# Control PoE port 3 of 8
disable = [
    "click/n/isShowPot3",
    "click/n/editPot3",
    "click/cls/poePortPwrTxt",
    "click/link/Disable",
    "click/n/submitPotedit",
]

enable = [
    "click/n/isShowPot3",
    "click/n/editPot3",
    "click/cls/poePortPwrTxt",
    "click/link/Enable",
    "click/n/submitPotedit",
]

logout = [
    "click/cls/src-views-header-nav-icon-button",
    "click/cls/icon-logout",
    "click/id/modal_footer_button_primary",
]


# sequence(login)
# sequence(disable)
# sequence(logout)



"""
massconfig.py
-------------

This module uses APISchema to serialize and deserialize the config
file, plus provides a schema for easy editing of the config.


The MaasConfig class plus SwitchDevice derived classes in the devices folder
provide an in memory representation of the contents of
the maaspower YAML configuration file that specifies the set of power
manager devices for which to serve web hooks.
"""

import re
from abc import ABC, abstractmethod
from collections.abc import Mapping, Sequence
from copy import deepcopy
from dataclasses import dataclass, fields
from typing import Annotated as A
from typing import Any, ClassVar

from apischema import deserialize, identity
from apischema.conversions import Conversion, deserializer
from typing_extensions import override

from .maas_globals import MaasResponse, T, desc


@dataclass(kw_only=True)
class SwitchDevice(ABC):
    """
    A base class for the switching devices that the webhook server will control.

    Concrete subclasses MUST provide a `type` field akin to this:

    type: Literal["ConreteDevice"] = "ConcreteDevice"

    Concrete subclasses are found in the devices subfolder
    """

    name: A[str, desc("A name for the switching device")]

    # command functions to be implemented in the derived classes
    @abstractmethod
    def turn_on(self) -> None: ...

    @abstractmethod
    def turn_off(self) -> None: ...

    @abstractmethod
    def query_state(self) -> str: ...

    def __post_init__(self):
        # allow regular expressions for names but if the name is an
        # illegal expression then just use it as a simple string match
        try:
            self._name_regx = re.compile(self.name)
        except re.error:
            self._name_regx = None

    # https://wyfo.github.io/apischema/examples/subclass_union/
    def __init_subclass__(cls):
        # Deserializers stack directly as a Union
        deserializer(Conversion(identity, source=cls, target=SwitchDevice))

    def copy(self, new_name: str, match) -> "SwitchDevice":
        """
        Create a copy of this device with a new name.
        All the fields of the object are reformatted with substitutions in
        regex matches using {name} for the whole match and {m1} {m2} etc
        for matching subgroups.

        This is used for creating a specific instance of a device from
        a regex defined device.
        """
        result = deepcopy(self)

        # TODO can't find an easy way to iterate over dataclass field instances
        for field in fields(self):
            if field.name == "name":
                continue
            setattr(result, field.name, match.expand(getattr(result, field.name)))

        result.name = new_name

        return result

    def do_command(self, command) -> str | None:
        result = None

        if command == "on":
            self.turn_on()
        elif command == "off":
            self.turn_off()
        elif command == "query":
            return self.query_state()
        else:
            raise ValueError("Illegal Command")
        return result


@dataclass(kw_only=True)
class RegexSwitchDevice(SwitchDevice, ABC):
    """
    An abstract `SwitchDevice` which has the ability to interpret reponses
    and convert them to the requisit MaasReponse values using regex.

    """

    query_on_regex: A[str, desc("match the on status return from query")] = "on"
    query_off_regex: A[str, desc("match the off status return from query")] = "off"

    @abstractmethod
    def run_query(self) -> str:
        """
        Ths method should be overridden by concrete classes. This method
        is called by query_state and it's response is run through query_regex_on
        and query_regex_off.

        returns: A value to be parsed by query_regex_on and query_regex_off.
        """
        ...

    @override
    def query_state(self) -> str:
        """
        Uses the regex patterns defined in query_on_regex and query_off_regex
        to ascertain the correct response.
        """
        query_response = self.run_query()
        if re.search(self.query_on_regex, query_response, flags=re.MULTILINE):
            return MaasResponse.on.value
        elif re.search(self.query_off_regex, query_response, flags=re.MULTILINE):
            return MaasResponse.off.value
        else:
            raise ValueError(
                f"Unknown power state response: \n{query_response}\n"
                f"\nfor regexes {self.query_on_regex}, {self.query_off_regex}"
            )


@dataclass
class MaasConfig:
    """
    Provides global information regarding webhook address, passwords etc.

    Plus a list of switch devices. The devices are generic in this module,
    config definitions and function for specific device types are
    in the devices folder.
    """

    name: A[str, desc("The name for this webhook server instance")]
    ip_address: A[str, desc("IP address to listen on")]
    port: A[int, desc("port to listen on")]
    username: A[str, desc("username for connecting to webhook")]
    password: A[str, desc("password for connecting to webhook")]

    devices: A[
        Sequence[SwitchDevice],
        desc("A list of the devices that this webhook server will control"),
    ]

    # this is a classvar to stop it appearing in the schema
    _devices: ClassVar[dict[str, SwitchDevice]] = {}

    @classmethod
    def deserialize(cls: type[T], d: Mapping[str, Any]) -> T:
        config: Any = deserialize(cls, d)
        # create indexed list of devices
        config._devices = {device.name: device for device in config.devices}
        return config

    def find_device(self, name: str):
        """
        use the indexed list to find the device or walk through the and check
        for regex matches.
        A regex match creates a new device which goes in the _devices cache
        so will not need matching a second time

        TODO https://github.com/gilesknap/maaspower/issues/10#issue-1222292918
        """
        if name in self._devices:
            return self._devices[name]
        for device in self.devices:
            if device._name_regx:
                match = device._name_regx.match(name)
                if match:
                    # create a copy with the correct name and substituted commands
                    # and cache it
                    self._devices[name] = device.copy(name, match)
                    return self._devices[name]
        return None



"""
smart_thing.py
--------------

Classes to represent the configuration and functionality for devices
that can be controlled via the SmartThings API.

See https://www.smartthings.com/
"""

import asyncio
from dataclasses import dataclass
from typing import Annotated as A
from typing import Literal

import aiohttp
from pysmartthings import SmartThings

from maaspower.maas_globals import desc
from maaspower.maasconfig import RegexSwitchDevice


@dataclass(kw_only=True)
class SmartThing(RegexSwitchDevice):
    """A device controlled via SmartThings"""

    on: A[str, desc("command line string to switch device on")]
    off: A[str, desc("command line string to switch device off")]
    query: A[str, desc("command line string to query device state")]
    query_on_regex: A[str, desc("match the on status return from query")] = "on"
    query_off_regex: A[str, desc("match the off status return from query")] = "off"

    type: Literal["SmartThingDevice"] = "SmartThingDevice"

    api_token: A[
        str,
        desc("SmartThings API token see https://account.smartthings.com/login"),
    ] = "none"

    device_id: A[
        str,
        desc("The SmartThings identifier for an individual device"),
    ] = "none"

    def turn_on(self):
        asyncio.run(self.switch(self.on))

    def turn_off(self):
        asyncio.run(self.switch(self.off))

    def run_query(self) -> str:
        return asyncio.run(self.switch(self.query, True))

    async def switch(self, cmd: str, query: bool = False):
        async with aiohttp.ClientSession() as session:
            result = ""

            api = SmartThings(session, self.api_token)
            devices = await api.devices()
            for device in devices:
                if device.device_id == self.device_id:
                    if query:
                        await device.status.refresh()
                        result = device.status.values.get(cmd)
                    else:
                        # commands are space separated sequences of identifiers
                        command = cmd.split(" ")
                        await device.command(*command)
                    break
            else:
                raise ValueError("Device not know to smartThings")

            return result



"""
kasa_device.py
--------------

This module contains the class to represent and control TP-Link Kasa smart devices via
the python-kasa API. It provides functionalities to turn the device on or off
and to query its current state.
"""

import asyncio
from dataclasses import dataclass, field
from typing import Annotated as A
from typing import Literal

from kasa import SmartDeviceException, SmartPlug

from maaspower.maas_globals import desc
from maaspower.maasconfig import SwitchDevice


@dataclass(kw_only=True)
class KasaDevice(SwitchDevice):
    ip_address: A[str, desc("The IP address of the Kasa device.")]
    type: Literal["KasaDevice"] = "KasaDevice"
    name: A[str, desc("The name of the switching device.")]
    plug: SmartPlug = field(
        init=False
    )  # The SmartPlug instance is initialized in __post_init__.

    def __post_init__(self):
        """Initialize the smart plug with the provided IP address
        after the object is instantiated."""
        super().__post_init__()
        self.plug = SmartPlug(self.ip_address)

    def turn_on(self) -> None:
        """Asynchronously turn the smart plug on using a coroutine
        managed by asyncio."""
        asyncio.run(self._turn_on())

    async def _turn_on(self):
        """Private coroutine to turn the smart plug on, catching
        and logging any exceptions."""
        try:
            await self.plug.update()
            await self.plug.turn_on()
        except SmartDeviceException as e:
            print(f"Failed to turn on the smart plug at {self.ip_address}: {str(e)}")

    def turn_off(self) -> None:
        """Asynchronously turn the smart plug off using a coroutine
        managed by asyncio."""
        asyncio.run(self._turn_off())

    async def _turn_off(self):
        """Private coroutine to turn the smart plug off, catching
        and logging any exceptions."""
        try:
            await self.plug.update()
            await self.plug.turn_off()
        except SmartDeviceException as e:
            print(f"Failed to turn off the smart plug at {self.ip_address}: {str(e)}")

    def query_state(self) -> str:
        """Asynchronously query the current state of the smart plug
        and return it as 'on' or 'off'."""
        return asyncio.run(self._query_state())

    async def _query_state(self) -> str:
        """Private coroutine to query the state of the smart plug,
        returning 'on', 'off', or 'error' if an exception occurs."""
        try:
            await self.plug.update()
            return "on" if self.plug.is_on else "off"
        except SmartDeviceException as e:
            print(
                f"Failed to query state of the smart plug \
                    at {self.ip_address}: {str(e)}"
            )
            return "error"



"""
maaspower package
=================
"""

from importlib.metadata import version

from . import maasconfig, webhook

__version__ = version("maaspower")
del version


# __all__ defines the public API for the package.
# Each module also defines its own __all__.
__all__ = ["__version__", "maasconfig", "webhook"]
